"""KanTu - Git-like image gallery management with delta encoding."""

__version__ = "1.0.1"
__author__ = "KanTu Developers"

from kantu.api import (
    ToolResult,
    add_image,
    export_image,
    find_similar,
    get_gallery_stats,
    get_image_info,
    init_gallery,
    list_images,
    remove_image,
)

__all__ = [
    "ToolResult",
    "init_gallery",
    "add_image",
    "remove_image",
    "get_image_info",
    "list_images",
    "export_image",
    "find_similar",
    "get_gallery_stats",
    "__version__",
]



"""PySide6 GUI for KanTu with native styling."""

import sys
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QFileDialog,
    QGroupBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QMainWindow,
    QMessageBox,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSplitter,
    QStatusBar,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from kantu.api import (
    add_image,
    export_image,
    get_gallery_stats,
    init_gallery,
    list_images,
    remove_image,
)
from kantu.core import KanTuCore


class MainWindow(QMainWindow):
    def __init__(self, gallery_path: str = "."):
        super().__init__()
        self.gallery_path = Path(gallery_path).resolve()
        self.setWindowTitle(f"KanTu - {self.gallery_path}")
        self.setMinimumSize(1100, 700)
        self.core = KanTuCore(str(self.gallery_path))
        self.setup_ui()
        self.refresh_gallery()

    def setup_ui(self):
        central = QWidget()
        self.setCentralWidget(central)
        main_layout = QVBoxLayout(central)
        main_layout.setContentsMargins(12, 12, 12, 12)
        main_layout.setSpacing(12)

        stats_group = QGroupBox("Gallery Statistics")
        stats_layout = QHBoxLayout(stats_group)
        stats_layout.setSpacing(20)
        self.stat_labels = {}
        for name, label in [
            ("Total", "Total Images"),
            ("Base", "Base Images"),
            ("Delta", "Delta Images"),
            ("Saved", "Storage Saved"),
        ]:
            container = QWidget()
            container_layout = QVBoxLayout(container)
            container_layout.setContentsMargins(0, 0, 0, 0)
            container_layout.setSpacing(2)
            value_label = QLabel("0")
            value_label.setStyleSheet("font-size: 18px; font-weight: bold;")
            name_label = QLabel(label)
            name_label.setStyleSheet("color: gray; font-size: 11px;")
            container_layout.addWidget(value_label)
            container_layout.addWidget(name_label)
            self.stat_labels[name] = value_label
            stats_layout.addWidget(container)
        stats_layout.addStretch()
        main_layout.addWidget(stats_group)

        toolbar_layout = QHBoxLayout()
        toolbar_layout.setSpacing(8)
        self.btn_init = QPushButton("Init Gallery")
        self.btn_init.clicked.connect(self.init_gallery_dialog)
        toolbar_layout.addWidget(self.btn_init)
        self.btn_add = QPushButton("Add Images")
        self.btn_add.clicked.connect(self.add_images_dialog)
        toolbar_layout.addWidget(self.btn_add)
        self.btn_refresh = QPushButton("Refresh")
        self.btn_refresh.clicked.connect(self.refresh_gallery)
        toolbar_layout.addWidget(self.btn_refresh)
        toolbar_layout.addStretch()
        self.progress = QProgressBar()
        self.progress.setVisible(False)
        self.progress.setFixedWidth(150)
        toolbar_layout.addWidget(self.progress)
        main_layout.addLayout(toolbar_layout)

        splitter = QSplitter(Qt.Horizontal)

        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        left_layout.setContentsMargins(0, 0, 0, 0)

        self.image_table = QTableWidget()
        self.image_table.setColumnCount(5)
        self.image_table.setHorizontalHeaderLabels(
            ["ID", "Type", "Dimensions", "Size", "Similarity"]
        )
        self.image_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.image_table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.image_table.setSelectionMode(QAbstractItemView.SingleSelection)
        self.image_table.setAlternatingRowColors(True)
        self.image_table.itemSelectionChanged.connect(self.on_selection_changed)
        left_layout.addWidget(self.image_table)

        action_layout = QHBoxLayout()
        self.btn_export = QPushButton("Export")
        self.btn_export.clicked.connect(self.export_selected)
        action_layout.addWidget(self.btn_export)
        self.btn_remove = QPushButton("Remove")
        self.btn_remove.clicked.connect(self.remove_selected)
        action_layout.addWidget(self.btn_remove)
        action_layout.addStretch()
        left_layout.addLayout(action_layout)

        splitter.addWidget(left_widget)

        right_widget = QWidget()
        right_layout = QVBoxLayout(right_widget)
        right_layout.setContentsMargins(0, 0, 0, 0)

        preview_group = QGroupBox("Preview")
        preview_layout = QVBoxLayout(preview_group)
        self.image_label = QLabel("Select an image")
        self.image_label.setAlignment(Qt.AlignCenter)
        self.image_label.setMinimumSize(300, 300)
        self.image_label.setStyleSheet("background-color: palette(base);")
        self.image_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        preview_layout.addWidget(self.image_label)
        right_layout.addWidget(preview_group)

        info_group = QGroupBox("Details")
        info_layout = QVBoxLayout(info_group)
        self.info_id = QLabel("ID: -")
        self.info_type = QLabel("Type: -")
        self.info_dim = QLabel("Dimensions: -")
        self.info_sim = QLabel("Similarity: -")
        for label in [self.info_id, self.info_type, self.info_dim, self.info_sim]:
            info_layout.addWidget(label)
        right_layout.addWidget(info_group)

        splitter.addWidget(right_widget)
        splitter.setSizes([600, 300])
        main_layout.addWidget(splitter)

        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

    def init_gallery_dialog(self):
        result = init_gallery(str(self.gallery_path))
        if result.success:
            QMessageBox.information(self, "Success", "Gallery initialized successfully")
            self.refresh_gallery()
        else:
            QMessageBox.warning(self, "Error", result.error or "Unknown error")

    def add_images_dialog(self):
        if not self.core.is_initialized():
            QMessageBox.warning(
                self, "Warning", "Gallery not initialized. Click 'Init Gallery' first."
            )
            return
        files, _ = QFileDialog.getOpenFileNames(
            self,
            "Select Images",
            "",
            "Images (*.png *.jpg *.jpeg *.bmp *.gif *.webp);;All Files (*)",
        )
        if files:
            self.progress.setVisible(True)
            self.progress.setMaximum(len(files))
            self.progress.setValue(0)
            for i, file_path in enumerate(files):
                result = add_image(file_path, str(self.gallery_path))
                if not result.success:
                    QMessageBox.warning(
                        self, "Error", f"Failed to add {Path(file_path).name}:\n{result.error}"
                    )
                self.progress.setValue(i + 1)
            self.progress.setVisible(False)
            self.refresh_gallery()

    def refresh_gallery(self):
        if not self.core.is_initialized():
            self.status_bar.showMessage("Not initialized")
            return
        result = list_images(str(self.gallery_path), limit=1000)
        if result.success:
            images = result.data.get("images", [])
            self.image_table.setRowCount(len(images))
            for row, img in enumerate(images):
                id_item = QTableWidgetItem(img["id"][:16])
                self.image_table.setItem(row, 0, id_item)
                type_item = QTableWidgetItem("Base" if img["is_base"] else "Delta")
                self.image_table.setItem(row, 1, type_item)
                self.image_table.setItem(
                    row, 2, QTableWidgetItem(f"{img['width']} × {img['height']}")
                )
                self.image_table.setItem(
                    row, 3, QTableWidgetItem(self._format_size(img["file_size"]))
                )
                sim = img.get("similarity_score", 1.0) if not img["is_base"] else 1.0
                self.image_table.setItem(row, 4, QTableWidgetItem(f"{sim:.1%}"))
        stats_result = get_gallery_stats(str(self.gallery_path))
        if stats_result.success:
            stats = stats_result.data
            self.stat_labels["Total"].setText(str(stats["total_images"]))
            self.stat_labels["Base"].setText(str(stats["base_images"]))
            self.stat_labels["Delta"].setText(str(stats["delta_images"]))
            self.stat_labels["Saved"].setText(f"{stats['savings_ratio']:.1%}")
            self.status_bar.showMessage(
                f"Total: {self._format_size(stats['total_size'])} | Saved: {stats['savings_ratio']:.1%}"
            )

    def on_selection_changed(self):
        selected = self.image_table.selectedItems()
        if not selected:
            self.clear_preview()
            return
        row = selected[0].row()
        id_item = self.image_table.item(row, 0)
        if id_item:
            self.show_image_preview(id_item.text())

    def show_image_preview(self, image_id: str):
        if not self.core.is_initialized():
            return
        arr = self.core.reconstruct_image(image_id)
        record = self.core.get_image_record(image_id)
        if arr is not None and record:
            h, w = arr.shape[:2]
            if len(arr.shape) == 3:
                ch = arr.shape[2]
                if ch == 3:
                    qimg = QImage(arr.data, w, h, 3 * w, QImage.Format_RGB888)
                elif ch == 4:
                    qimg = QImage(arr.data, w, h, 4 * w, QImage.Format_RGBA8888)
                else:
                    qimg = QImage(arr.data, w, h, w, QImage.Format_Grayscale8)
            else:
                qimg = QImage(arr.data, w, h, w, QImage.Format_Grayscale8)
            pixmap = QPixmap.fromImage(qimg)
            scaled = pixmap.scaled(
                self.image_label.size(), Qt.KeepAspectRatio, Qt.SmoothTransformation
            )
            self.image_label.setPixmap(scaled)
            self.info_id.setText(f"ID: {image_id[:16]}...")
            self.info_type.setText(f"Type: {'Base' if record['is_base'] else 'Delta'}")
            self.info_dim.setText(f"Dimensions: {record['width']} × {record['height']}")
            sim = record.get("similarity_score", 1.0) if not record["is_base"] else 1.0
            self.info_sim.setText(f"Similarity: {sim:.1%}")
        else:
            self.clear_preview()

    def clear_preview(self):
        self.image_label.clear()
        self.image_label.setText("Select an image")
        self.info_id.setText("ID: -")
        self.info_type.setText("Type: -")
        self.info_dim.setText("Dimensions: -")
        self.info_sim.setText("Similarity: -")

    def export_selected(self):
        selected = self.image_table.selectedItems()
        if not selected:
            QMessageBox.warning(self, "Warning", "No image selected")
            return
        row = selected[0].row()
        id_item = self.image_table.item(row, 0)
        if not id_item:
            return
        image_id = id_item.text()
        file_path, _ = QFileDialog.getSaveFileName(
            self, "Export Image", f"{image_id}.png", "PNG (*.png);;JPEG (*.jpg)"
        )
        if file_path:
            result = export_image(image_id, file_path, str(self.gallery_path))
            if result.success:
                QMessageBox.information(self, "Success", f"Exported to {file_path}")
            else:
                QMessageBox.warning(self, "Error", result.error or "Export failed")

    def remove_selected(self):
        selected = self.image_table.selectedItems()
        if not selected:
            QMessageBox.warning(self, "Warning", "No image selected")
            return
        row = selected[0].row()
        id_item = self.image_table.item(row, 0)
        if not id_item:
            return
        image_id = id_item.text()
        reply = QMessageBox.question(
            self, "Confirm", f"Remove image {image_id[:16]}...?", QMessageBox.Yes | QMessageBox.No
        )
        if reply == QMessageBox.Yes:
            result = remove_image(image_id, str(self.gallery_path))
            if result.success:
                self.refresh_gallery()
                self.clear_preview()
            else:
                QMessageBox.warning(self, "Error", result.error or "Remove failed")

    def _format_size(self, size: int) -> str:
        s = float(size)
        for unit in ["B", "KB", "MB", "GB"]:
            if s < 1024:
                return f"{s:.1f} {unit}"
            s /= 1024
        return f"{s:.1f} TB"

    def closeEvent(self, event):
        self.core.close()
        event.accept()


def run_gui(gallery_path: str = ".", no_web: bool = False):
    app = QApplication(sys.argv)
    app.setApplicationName("KanTu")
    window = MainWindow(gallery_path)
    window.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    run_gui()



"""Test configuration and fixtures for KanTu."""

import tempfile
from pathlib import Path

import numpy as np
import pytest
from PIL import Image


@pytest.fixture
def temp_gallery():
    """Create a temporary gallery directory for testing."""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield tmpdir


@pytest.fixture
def sample_images(temp_gallery):
    """Create sample test images."""
    images_dir = Path(temp_gallery) / "images"
    images_dir.mkdir()
    base_img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
    base_path = images_dir / "base.png"
    Image.fromarray(base_img).save(base_path)
    similar_img = base_img.copy()
    similar_img[10:20, 10:20] = 255 - similar_img[10:20, 10:20]
    similar_path = images_dir / "similar.png"
    Image.fromarray(similar_img).save(similar_path)
    diff_img = np.random.randint(0, 255, (100, 100, 3), dtype=np.uint8)
    diff_path = images_dir / "different.png"
    Image.fromarray(diff_img).save(diff_path)
    return {
        "base": str(base_path),
        "similar": str(similar_path),
        "different": str(diff_path),
        "dir": str(images_dir),
    }


@pytest.fixture
def initialized_gallery(temp_gallery, sample_images):
    """Create an initialized gallery with one base image."""
    from kantu.api import add_image, init_gallery

    init_gallery(temp_gallery)
    add_image(sample_images["base"], temp_gallery, force_base=True)
    return temp_gallery



"""Command-line interface for KanTu."""

import argparse
import json
import sys

from kantu import __version__
from kantu.api import (
    ToolResult,
    add_image,
    export_image,
    find_similar,
    get_gallery_stats,
    get_image_info,
    init_gallery,
    list_images,
    remove_image,
    set_config,
)


def print_result(result: ToolResult, json_output: bool = False) -> None:
    if json_output:
        print(json.dumps(result.to_dict(), indent=2, default=str))
    elif result.success:
        if result.data:
            if isinstance(result.data, dict):
                for key, value in result.data.items():
                    if key == "images":
                        continue
                    print(f"{key}: {value}")
            else:
                print(result.data)
        if result.metadata.get("message"):
            print(result.metadata["message"])
    else:
        print(f"Error: {result.error}", file=sys.stderr)
        sys.exit(1)


def cmd_init(args: argparse.Namespace) -> None:
    result = init_gallery(args.path)
    print_result(result, getattr(args, "json_output", False))


def cmd_add(args: argparse.Namespace) -> None:
    for image_path in args.images:
        result = add_image(
            image_path=image_path,
            gallery_path=args.gallery,
            similarity_threshold=args.threshold,
            force_base=args.force_base,
        )
        print(f"\n{image_path}:")
        print_result(result, getattr(args, "json_output", False))


def cmd_remove(args: argparse.Namespace) -> None:
    result = remove_image(args.image_id, args.gallery)
    print_result(result, getattr(args, "json_output", False))


def cmd_info(args: argparse.Namespace) -> None:
    result = get_image_info(args.image_id, args.gallery)
    print_result(result, getattr(args, "json_output", False))


def cmd_list(args: argparse.Namespace) -> None:
    result = list_images(args.gallery, args.limit, args.offset)
    json_output = getattr(args, "json_output", False)
    if json_output:
        print_result(result, json_output)
    elif result.success:
        images = result.data.get("images", [])
        print(f"Total images: {result.data.get('total', 0)}")
        print("-" * 60)
        for img in images:
            img_type = "base" if img["is_base"] else "delta"
            print(
                f"ID: {img['id'][:12]}... | Type: {img_type} | Size: {img['width']}x{img['height']}"
            )
            if not img["is_base"] and img.get("base_id"):
                print(
                    f"  Base: {img['base_id'][:12]}... | Similarity: {img.get('similarity_score', 0):.2%}"
                )


def cmd_export(args: argparse.Namespace) -> None:
    result = export_image(args.image_id, args.output, args.gallery)
    print_result(result, getattr(args, "json_output", False))


def cmd_similar(args: argparse.Namespace) -> None:
    result = find_similar(args.image, args.gallery, args.threshold)
    json_output = getattr(args, "json_output", False)
    if json_output:
        print_result(result, json_output)
    elif result.success:
        print(f"Image hash: {result.data.get('phash')}")
        similar = result.data.get("similar_images", [])
        if similar:
            print(f"\nFound {len(similar)} similar images:")
            for img in similar:
                print(f"  {img['id'][:12]}... | Distance: {img['distance']}")
        else:
            print("No similar images found.")


def cmd_stats(args: argparse.Namespace) -> None:
    result = get_gallery_stats(args.gallery)
    json_output = getattr(args, "json_output", False)
    if json_output:
        print_result(result, json_output)
    elif result.success:
        stats = result.data
        print("Gallery Statistics")
        print("=" * 40)
        print(f"Total images:    {stats['total_images']}")
        print(f"Base images:     {stats['base_images']}")
        print(f"Delta images:    {stats['delta_images']}")
        print("-" * 40)
        print(f"Base size:       {_format_size(stats['base_size'])}")
        print(f"Delta size:      {_format_size(stats['delta_size'])}")
        print(f"Total size:      {_format_size(stats['total_size'])}")
        print("-" * 40)
        print(f"Original size:   {_format_size(stats['original_size'])}")
        print(f"Storage saved:   {stats['savings_ratio']:.1%}")


def cmd_config(args: argparse.Namespace) -> None:
    kwargs = {}
    if args.similarity_threshold is not None:
        kwargs["similarity_threshold"] = args.similarity_threshold
    if args.min_delta_ratio is not None:
        kwargs["min_delta_ratio"] = args.min_delta_ratio
    if args.max_hamming_distance is not None:
        kwargs["max_hamming_distance"] = args.max_hamming_distance
    result = set_config(args.gallery, **kwargs)
    print_result(result, args.json)


def cmd_gui(args: argparse.Namespace) -> None:
    from kantu.gui import run_gui

    run_gui(args.gallery, args.no_web)


def cmd_web(args: argparse.Namespace) -> None:
    from kantu.app import run_server

    run_server(host=args.host, port=args.port, gallery_path=args.gallery)


def _format_size(size: int) -> str:
    s = float(size)
    for unit in ["B", "KB", "MB", "GB"]:
        if s < 1024:
            return f"{s:.2f} {unit}"
        s /= 1024
    return f"{s:.2f} TB"


def main() -> int:
    parser = argparse.ArgumentParser(
        prog="kantu",
        description="Git-like image gallery management with delta encoding",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("-V", "--version", action="version", version=f"kantu {__version__}")
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add_argument("--json", action="store_true", dest="json_output", help="Output as JSON")
    parser.add_argument("-q", "--quiet", action="store_true", help="Suppress output")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")
    init_parser = subparsers.add_parser("init", help="Initialize a new gallery")
    init_parser.add_argument("path", nargs="?", default=".", help="Gallery path")
    init_parser.set_defaults(func=cmd_init)
    add_parser = subparsers.add_parser("add", help="Add images to gallery")
    add_parser.add_argument("images", nargs="+", help="Image files to add")
    add_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    add_parser.add_argument("-t", "--threshold", type=float, help="Similarity threshold (0-1)")
    add_parser.add_argument("--force-base", action="store_true", help="Force store as base image")
    add_parser.set_defaults(func=cmd_add)
    remove_parser = subparsers.add_parser(
        "remove", aliases=["rm"], help="Remove image from gallery"
    )
    remove_parser.add_argument("image_id", help="Image ID to remove")
    remove_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    remove_parser.set_defaults(func=cmd_remove)
    info_parser = subparsers.add_parser("info", help="Get image information")
    info_parser.add_argument("image_id", help="Image ID")
    info_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    info_parser.set_defaults(func=cmd_info)
    list_parser = subparsers.add_parser("list", aliases=["ls"], help="List images in gallery")
    list_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    list_parser.add_argument("-l", "--limit", type=int, default=100, help="Limit results")
    list_parser.add_argument("-o", "--offset", type=int, default=0, help="Offset results")
    list_parser.set_defaults(func=cmd_list)
    export_parser = subparsers.add_parser("export", help="Export image from gallery")
    export_parser.add_argument("image_id", help="Image ID to export")
    export_parser.add_argument("-o", "--output", required=True, help="Output file path")
    export_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    export_parser.set_defaults(func=cmd_export)
    similar_parser = subparsers.add_parser("similar", help="Find similar images")
    similar_parser.add_argument("image", help="Image file to compare")
    similar_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    similar_parser.add_argument(
        "-t", "--threshold", type=int, default=10, help="Hamming distance threshold"
    )
    similar_parser.set_defaults(func=cmd_similar)
    stats_parser = subparsers.add_parser("stats", help="Show gallery statistics")
    stats_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    stats_parser.set_defaults(func=cmd_stats)
    config_parser = subparsers.add_parser("config", help="Configure gallery settings")
    config_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    config_parser.add_argument(
        "--similarity-threshold", type=float, help="Similarity threshold (0-1)"
    )
    config_parser.add_argument(
        "--min-delta-ratio", type=float, help="Minimum delta ratio to store as delta"
    )
    config_parser.add_argument(
        "--max-hamming-distance", type=int, help="Maximum hamming distance for similarity"
    )
    config_parser.set_defaults(func=cmd_config)
    gui_parser = subparsers.add_parser("gui", help="Launch GUI")
    gui_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    gui_parser.add_argument("--no-web", action="store_true", help="Disable embedded web server")
    gui_parser.set_defaults(func=cmd_gui)
    web_parser = subparsers.add_parser("web", help="Launch web server")
    web_parser.add_argument("--host", default="127.0.0.1", help="Host address")
    web_parser.add_argument("--port", type=int, default=5000, help="Port number")
    web_parser.add_argument("-g", "--gallery", default=".", help="Gallery path")
    web_parser.set_defaults(func=cmd_web)
    args = parser.parse_args()
    if args.command is None:
        from kantu.gui import run_gui

        run_gui(".", False)
        return 0
    args.func(args)
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Unified API for KanTu with ToolResult pattern."""

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional

from kantu.core import DeltaType, ImageRecord, KanTuCore


@dataclass
class ToolResult:
    success: bool
    data: Any = None
    error: Optional[str] = None
    metadata: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            "success": self.success,
            "data": self.data,
            "error": self.error,
            "metadata": self.metadata,
        }


def init_gallery(path: str = ".") -> ToolResult:
    try:
        core = KanTuCore(path)
        if core.is_initialized():
            return ToolResult(
                success=True,
                data={"path": str(core.gallery_path)},
                error=None,
                metadata={"message": "Gallery already initialized"},
            )
        core.init_gallery()
        return ToolResult(
            success=True,
            data={"path": str(core.gallery_path)},
            error=None,
            metadata={"message": "Gallery initialized successfully"},
        )
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def add_image(
    image_path: str,
    gallery_path: str = ".",
    similarity_threshold: Optional[float] = None,
    force_base: bool = False,
) -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(
                success=False, error="Gallery not initialized. Run 'kantu init' first."
            )
        core._load_config()
        if similarity_threshold is not None:
            core.config.similarity_threshold = similarity_threshold
        if not Path(image_path).exists():
            return ToolResult(success=False, error=f"Image not found: {image_path}")
        image_id = core.compute_id(image_path)
        existing = core.get_image_record(image_id)
        if existing:
            return ToolResult(
                success=True,
                data={"id": image_id, "status": "already_exists"},
                metadata={"message": "Image already in gallery"},
            )
        phash = core.compute_hash(image_path)
        width, height = core.get_image_dimensions(image_path)
        original_size = core.get_file_size(image_path)
        if force_base:
            base_path = core.save_base_image(image_path, image_id)
            record = ImageRecord(
                id=image_id,
                phash=phash,
                original_path=str(Path(image_path).resolve()),
                is_base=True,
                width=width,
                height=height,
                file_size=original_size,
            )
            core.add_image_record(record)
            return ToolResult(
                success=True,
                data={"id": image_id, "type": "base", "path": base_path},
                metadata={"message": "Image added as base (forced)"},
            )
        similar = core.find_similar_images(phash)
        best_match = None
        best_ssim = 0
        for candidate in similar:
            if candidate["id"] == image_id:
                continue
            ssim_score = core.compute_ssim(image_path, candidate["original_path"])
            if ssim_score > best_ssim and ssim_score >= core.config.similarity_threshold:
                best_ssim = ssim_score
                best_match = candidate
        if best_match:
            delta, target_size = core.compute_pixel_delta(best_match["original_path"], image_path)
            delta_compressed = core.save_delta(delta, image_id)
            delta_size = core.get_file_size(delta_compressed)
            savings_ratio = 1 - (delta_size / original_size)
            if savings_ratio >= core.config.min_delta_ratio:
                record = ImageRecord(
                    id=image_id,
                    phash=phash,
                    original_path=str(Path(image_path).resolve()),
                    is_base=False,
                    base_id=best_match["id"],
                    delta_type=DeltaType.PIXEL,
                    delta_path=delta_compressed,
                    width=width,
                    height=height,
                    file_size=delta_size,
                    similarity_score=best_ssim,
                )
                core.add_image_record(record)
                return ToolResult(
                    success=True,
                    data={
                        "id": image_id,
                        "type": "delta",
                        "base_id": best_match["id"],
                        "similarity": best_ssim,
                        "savings_ratio": savings_ratio,
                    },
                    metadata={"message": f"Image stored as delta (saved {savings_ratio:.1%})"},
                )
        base_path = core.save_base_image(image_path, image_id)
        record = ImageRecord(
            id=image_id,
            phash=phash,
            original_path=str(Path(image_path).resolve()),
            is_base=True,
            width=width,
            height=height,
            file_size=original_size,
        )
        core.add_image_record(record)
        return ToolResult(
            success=True,
            data={"id": image_id, "type": "base", "path": base_path},
            metadata={"message": "Image added as base (no similar images found)"},
        )
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def remove_image(image_id: str, gallery_path: str = ".") -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        record = core.get_image_record(image_id)
        if not record:
            return ToolResult(success=False, error=f"Image not found: {image_id}")
        conn = core._get_connection()
        cursor = conn.cursor()
        cursor.execute(
            "SELECT COUNT(*) FROM images WHERE base_id = ? AND id != ?",
            (image_id, image_id),
        )
        dependent_count = cursor.fetchone()[0]
        if dependent_count > 0:
            return ToolResult(
                success=False,
                error=f"Cannot remove: {dependent_count} images depend on this base image",
            )
        core.delete_image_record(image_id)
        return ToolResult(
            success=True,
            data={"id": image_id},
            metadata={"message": "Image removed successfully"},
        )
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def get_image_info(image_id: str, gallery_path: str = ".") -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        record = core.get_image_record(image_id)
        if not record:
            return ToolResult(success=False, error=f"Image not found: {image_id}")
        return ToolResult(success=True, data=record)
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def list_images(gallery_path: str = ".", limit: int = 100, offset: int = 0) -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        images = core.list_all_images()
        paginated = images[offset : offset + limit]
        return ToolResult(
            success=True,
            data={"images": paginated, "total": len(images), "limit": limit, "offset": offset},
        )
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def export_image(image_id: str, output_path: str, gallery_path: str = ".") -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        success = core.export_image(image_id, output_path)
        if success:
            return ToolResult(
                success=True,
                data={"output_path": output_path},
                metadata={"message": "Image exported successfully"},
            )
        return ToolResult(success=False, error="Failed to export image")
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def find_similar(image_path: str, gallery_path: str = ".", threshold: int = 10) -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        if not Path(image_path).exists():
            return ToolResult(success=False, error=f"Image not found: {image_path}")
        phash = core.compute_hash(image_path)
        similar = core.find_similar_images(phash, threshold)
        return ToolResult(success=True, data={"phash": phash, "similar_images": similar})
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def get_gallery_stats(gallery_path: str = ".") -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        stats = core.get_gallery_stats()
        return ToolResult(success=True, data=stats)
    except Exception as e:
        return ToolResult(success=False, error=str(e))


def set_config(
    gallery_path: str = ".",
    similarity_threshold: Optional[float] = None,
    min_delta_ratio: Optional[float] = None,
    max_hamming_distance: Optional[int] = None,
) -> ToolResult:
    try:
        core = KanTuCore(gallery_path)
        if not core.is_initialized():
            return ToolResult(success=False, error="Gallery not initialized")
        core._load_config()
        changes = {}
        if similarity_threshold is not None:
            core.config.similarity_threshold = similarity_threshold
            changes["similarity_threshold"] = similarity_threshold
        if min_delta_ratio is not None:
            core.config.min_delta_ratio = min_delta_ratio
            changes["min_delta_ratio"] = min_delta_ratio
        if max_hamming_distance is not None:
            core.config.max_hamming_distance = max_hamming_distance
            changes["max_hamming_distance"] = max_hamming_distance
        core._save_config()
        return ToolResult(
            success=True,
            data={"changes": changes},
            metadata={"message": "Configuration updated"},
        )
    except Exception as e:
        return ToolResult(success=False, error=str(e))



"""Flask web application for KanTu."""

import io
import os
from pathlib import Path

from flask import Flask, jsonify, render_template_string, request, send_file

from kantu.api import (
    ToolResult,
    add_image,
    export_image,
    find_similar,
    get_gallery_stats,
    get_image_info,
    init_gallery,
    list_images,
    remove_image,
    set_config,
)
from kantu.core import KanTuCore

app = Flask(__name__)
app.config["MAX_CONTENT_LENGTH"] = 100 * 1024 * 1024
GALLERY_PATH = "."


@app.route("/")
def index():
    return render_template_string(INDEX_HTML)


@app.route("/api/init", methods=["POST"])
def api_init():
    result = init_gallery(GALLERY_PATH)
    return jsonify(result.to_dict())


@app.route("/api/images", methods=["GET"])
def api_list_images():
    limit = request.args.get("limit", 100, type=int)
    offset = request.args.get("offset", 0, type=int)
    result = list_images(GALLERY_PATH, limit, offset)
    return jsonify(result.to_dict())


@app.route("/api/images", methods=["POST"])
def api_add_image():
    if "file" not in request.files:
        return jsonify(ToolResult(False, error="No file provided").to_dict()), 400
    file = request.files["file"]
    if file.filename == "":
        return jsonify(ToolResult(False, error="No file selected").to_dict()), 400
    temp_dir = Path(GALLERY_PATH) / ".kantu" / "temp"
    temp_dir.mkdir(parents=True, exist_ok=True)
    temp_path = temp_dir / file.filename
    file.save(temp_path)
    threshold = request.form.get("threshold", type=float)
    force_base = request.form.get("force_base", "false").lower() == "true"
    result = add_image(
        str(temp_path), GALLERY_PATH, similarity_threshold=threshold, force_base=force_base
    )
    try:
        os.remove(temp_path)
    except Exception:
        pass
    return jsonify(result.to_dict())


@app.route("/api/images/<image_id>", methods=["GET"])
def api_get_image(image_id):
    result = get_image_info(image_id, GALLERY_PATH)
    return jsonify(result.to_dict())


@app.route("/api/images/<image_id>", methods=["DELETE"])
def api_delete_image(image_id):
    result = remove_image(image_id, GALLERY_PATH)
    return jsonify(result.to_dict())


@app.route("/api/images/<image_id>/export", methods=["GET"])
def api_export_image(image_id):
    output_dir = Path(GALLERY_PATH) / ".kantu" / "temp"
    output_dir.mkdir(parents=True, exist_ok=True)
    output_path = output_dir / f"{image_id}.png"
    result = export_image(image_id, str(output_path), GALLERY_PATH)
    if result.success:
        return send_file(output_path, as_attachment=True, download_name=f"{image_id}.png")
    return jsonify(result.to_dict()), 400


@app.route("/api/images/<image_id>/preview", methods=["GET"])
def api_preview_image(image_id):
    core = KanTuCore(GALLERY_PATH)
    if not core.is_initialized():
        return jsonify(ToolResult(False, error="Gallery not initialized").to_dict()), 400
    arr = core.reconstruct_image(image_id)
    if arr is None:
        return jsonify(ToolResult(False, error="Image not found").to_dict()), 404
    from PIL import Image

    img = Image.fromarray(arr)
    img_io = io.BytesIO()
    img.save(img_io, "PNG")
    img_io.seek(0)
    return send_file(img_io, mimetype="image/png")


@app.route("/api/similar", methods=["POST"])
def api_find_similar():
    if "file" not in request.files:
        return jsonify(ToolResult(False, error="No file provided").to_dict()), 400
    file = request.files["file"]
    temp_dir = Path(GALLERY_PATH) / ".kantu" / "temp"
    temp_dir.mkdir(parents=True, exist_ok=True)
    temp_path = temp_dir / file.filename
    file.save(temp_path)
    threshold = request.form.get("threshold", 10, type=int)
    result = find_similar(str(temp_path), GALLERY_PATH, threshold)
    try:
        os.remove(temp_path)
    except Exception:
        pass
    return jsonify(result.to_dict())


@app.route("/api/stats", methods=["GET"])
def api_stats():
    result = get_gallery_stats(GALLERY_PATH)
    return jsonify(result.to_dict())


@app.route("/api/config", methods=["GET", "PUT"])
def api_config():
    if request.method == "GET":
        core = KanTuCore(GALLERY_PATH)
        if not core.is_initialized():
            return jsonify(ToolResult(False, error="Gallery not initialized").to_dict()), 400
        core._load_config()
        return jsonify(
            ToolResult(
                True,
                data={
                    "similarity_threshold": core.config.similarity_threshold,
                    "min_delta_ratio": core.config.min_delta_ratio,
                    "max_hamming_distance": core.config.max_hamming_distance,
                },
            ).to_dict()
        )
    else:
        data = request.json or {}
        result = set_config(
            GALLERY_PATH,
            similarity_threshold=data.get("similarity_threshold"),
            min_delta_ratio=data.get("min_delta_ratio"),
            max_hamming_distance=data.get("max_hamming_distance"),
        )
        return jsonify(result.to_dict())


def run_server(host: str = "127.0.0.1", port: int = 5000, gallery_path: str = "."):
    global GALLERY_PATH
    GALLERY_PATH = gallery_path
    print(f"Starting KanTu web server at http://{host}:{port}")
    print(f"Gallery path: {Path(gallery_path).resolve()}")
    app.run(host=host, port=port, debug=False)


INDEX_HTML = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>KanTu - Image Gallery Manager</title>
    <style>
        body { font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif; margin: 0; padding: 20px; background: #fff; color: #333; }
        h1 { margin: 0 0 20px 0; font-size: 24px; }
        .stats { display: flex; gap: 24px; margin-bottom: 20px; padding: 16px; background: #f5f5f5; border-radius: 8px; }
        .stat { text-align: center; }
        .stat-value { font-size: 24px; font-weight: bold; }
        .stat-label { font-size: 12px; color: #666; }
        .toolbar { margin-bottom: 20px; display: flex; gap: 8px; flex-wrap: wrap; }
        button { padding: 8px 16px; border: 1px solid #ccc; border-radius: 4px; background: #fff; cursor: pointer; font-size: 14px; }
        button:hover { background: #f0f0f0; }
        button.primary { background: #0066cc; color: #fff; border-color: #0066cc; }
        button.primary:hover { background: #0055aa; }
        button.danger { background: #cc3333; color: #fff; border-color: #cc3333; }
        .gallery { display: grid; grid-template-columns: repeat(auto-fill, minmax(200px, 1fr)); gap: 16px; }
        .card { border: 1px solid #ddd; border-radius: 8px; overflow: hidden; cursor: pointer; }
        .card:hover { box-shadow: 0 2px 8px rgba(0,0,0,0.1); }
        .card img { width: 100%; height: 150px; object-fit: cover; background: #f5f5f5; }
        .card-body { padding: 12px; }
        .card-id { font-family: monospace; font-size: 12px; color: #666; margin-bottom: 4px; }
        .badge { display: inline-block; padding: 2px 8px; border-radius: 4px; font-size: 11px; font-weight: 500; }
        .badge.base { background: #e6f3e6; color: #006600; }
        .badge.delta { background: #fff3e6; color: #cc6600; }
        .modal { display: none; position: fixed; top: 0; left: 0; width: 100%; height: 100%; background: rgba(0,0,0,0.5); justify-content: center; align-items: center; }
        .modal.active { display: flex; }
        .modal-content { background: #fff; padding: 24px; border-radius: 8px; max-width: 700px; width: 90%; }
        .modal-header { display: flex; justify-content: space-between; align-items: center; margin-bottom: 16px; }
        .modal-header h2 { margin: 0; font-size: 18px; }
        .modal-close { background: none; border: none; font-size: 24px; cursor: pointer; color: #999; }
        .modal-body { display: flex; gap: 16px; }
        .modal-image { flex: 1; text-align: center; }
        .modal-image img { max-width: 100%; max-height: 300px; }
        .modal-info { width: 200px; }
        .modal-info p { margin: 0 0 8px 0; }
        .modal-info label { font-size: 11px; color: #666; text-transform: uppercase; }
        .modal-actions { margin-top: 16px; display: flex; gap: 8px; }
        input[type="file"] { display: none; }
        .empty { text-align: center; padding: 40px; color: #666; }
    </style>
</head>
<body>
    <h1>KanTu</h1>
    <div class="stats">
        <div class="stat"><div class="stat-value" id="stat-total">0</div><div class="stat-label">Total</div></div>
        <div class="stat"><div class="stat-value" id="stat-base">0</div><div class="stat-label">Base</div></div>
        <div class="stat"><div class="stat-value" id="stat-delta">0</div><div class="stat-label">Delta</div></div>
        <div class="stat"><div class="stat-value" id="stat-saved">0%</div><div class="stat-label">Saved</div></div>
    </div>
    <div class="toolbar">
        <button class="primary" onclick="initGallery()">Init Gallery</button>
        <input type="file" id="file-input" multiple accept="image/*" onchange="uploadFiles(this.files)">
        <button class="primary" onclick="document.getElementById('file-input').click()">Add Images</button>
        <button onclick="refreshGallery()">Refresh</button>
    </div>
    <div class="gallery" id="gallery"></div>
    <div class="modal" id="modal">
        <div class="modal-content">
            <div class="modal-header">
                <h2>Image Preview</h2>
                <button class="modal-close" onclick="closeModal()">&times;</button>
            </div>
            <div class="modal-body">
                <div class="modal-image"><img id="modal-img" src="" alt="Preview"></div>
                <div class="modal-info">
                    <p><label>ID</label><br><span id="modal-id">-</span></p>
                    <p><label>Type</label><br><span id="modal-type">-</span></p>
                    <p><label>Dimensions</label><br><span id="modal-dim">-</span></p>
                    <p><label>Similarity</label><br><span id="modal-sim">-</span></p>
                </div>
            </div>
            <div class="modal-actions">
                <button onclick="closeModal()">Close</button>
                <button class="primary" onclick="exportImage()">Export</button>
                <button class="danger" onclick="deleteImage()">Delete</button>
            </div>
        </div>
    </div>
    <script>
        let currentImageId = null;
        async function fetchAPI(endpoint, options = {}) {
            const response = await fetch('/api' + endpoint, options);
            return response.json();
        }
        async function refreshGallery() {
            const [imagesRes, statsRes] = await Promise.all([fetchAPI('/images?limit=1000'), fetchAPI('/stats')]);
            if (imagesRes.success) renderGallery(imagesRes.data.images);
            if (statsRes.success) updateStats(statsRes.data);
        }
        function updateStats(stats) {
            document.getElementById('stat-total').textContent = stats.total_images;
            document.getElementById('stat-base').textContent = stats.base_images;
            document.getElementById('stat-delta').textContent = stats.delta_images;
            document.getElementById('stat-saved').textContent = (stats.savings_ratio * 100).toFixed(1) + '%';
        }
        function renderGallery(images) {
            const gallery = document.getElementById('gallery');
            if (!images.length) {
                gallery.innerHTML = '<div class="empty">No images. Initialize gallery and add images to get started.</div>';
                return;
            }
            gallery.innerHTML = images.map(img => `
                <div class="card" onclick="showImage('${img.id}')">
                    <img src="/api/images/${img.id}/preview" alt="${img.id}" onerror="this.src='data:image/svg+xml,<svg xmlns=%22http://www.w3.org/2000/svg%22 viewBox=%220 0 200 150%22><rect fill=%22%23f5f5f5%22 width=%22200%22 height=%22150%22/><text x=%22100%22 y=%2275%22 text-anchor=%22middle%22 fill=%22%23999%22>No Preview</text></svg>'">
                    <div class="card-body">
                        <div class="card-id">${img.id.substring(0, 12)}...</div>
                        <span class="badge ${img.is_base ? 'base' : 'delta'}">${img.is_base ? 'Base' : 'Delta'}</span>
                    </div>
                </div>
            `).join('');
        }
        async function initGallery() {
            const result = await fetchAPI('/init', { method: 'POST' });
            alert(result.success ? 'Gallery initialized!' : 'Error: ' + result.error);
            refreshGallery();
        }
        async function uploadFiles(files) {
            for (const file of files) {
                const formData = new FormData();
                formData.append('file', file);
                await fetchAPI('/images', { method: 'POST', body: formData });
            }
            refreshGallery();
        }
        async function showImage(id) {
            currentImageId = id;
            const result = await fetchAPI(`/images/${id}`);
            if (result.success) {
                document.getElementById('modal-img').src = `/api/images/${id}/preview`;
                document.getElementById('modal-id').textContent = id.substring(0, 16) + '...';
                document.getElementById('modal-type').textContent = result.data.is_base ? 'Base' : 'Delta';
                document.getElementById('modal-dim').textContent = `${result.data.width} × ${result.data.height}`;
                document.getElementById('modal-sim').textContent = result.data.is_base ? 'N/A' : (result.data.similarity_score * 100).toFixed(1) + '%';
                document.getElementById('modal').classList.add('active');
            }
        }
        function closeModal() { document.getElementById('modal').classList.remove('active'); currentImageId = null; }
        function exportImage() { if (currentImageId) window.location.href = `/api/images/${currentImageId}/export`; }
        async function deleteImage() {
            if (currentImageId && confirm('Delete this image?')) {
                const result = await fetchAPI(`/images/${currentImageId}`, { method: 'DELETE' });
                alert(result.success ? 'Deleted!' : 'Error: ' + result.error);
                closeModal();
                refreshGallery();
            }
        }
        document.addEventListener('keydown', e => { if (e.key === 'Escape') closeModal(); });
        refreshGallery();
    </script>
</body>
</html>"""


if __name__ == "__main__":
    run_server()



"""Tests for KanTu API functions."""

from kantu.api import (
    ToolResult,
    add_image,
    export_image,
    find_similar,
    get_gallery_stats,
    get_image_info,
    init_gallery,
    list_images,
    remove_image,
    set_config,
)


class TestToolResult:
    def test_success_result(self):
        result = ToolResult(success=True, data={"key": "value"})
        assert result.success
        assert result.data == {"key": "value"}
        assert result.error is None

    def test_error_result(self):
        result = ToolResult(success=False, error="Something went wrong")
        assert not result.success
        assert result.error == "Something went wrong"

    def test_to_dict(self):
        result = ToolResult(
            success=True,
            data={"id": 123},
            metadata={"message": "Done"},
        )
        d = result.to_dict()
        assert d["success"]
        assert d["data"]["id"] == 123
        assert d["metadata"]["message"] == "Done"


class TestInitGallery:
    def test_init_new_gallery(self, temp_gallery):
        result = init_gallery(temp_gallery)
        assert result.success
        assert "path" in result.data

    def test_init_existing_gallery(self, temp_gallery):
        init_gallery(temp_gallery)
        result = init_gallery(temp_gallery)
        assert result.success
        assert "already initialized" in result.metadata.get("message", "")


class TestAddImage:
    def test_add_base_image(self, temp_gallery, sample_images):
        init_gallery(temp_gallery)
        result = add_image(sample_images["base"], temp_gallery, force_base=True)
        assert result.success
        assert result.data["type"] == "base"

    def test_add_duplicate_image(self, temp_gallery, sample_images):
        init_gallery(temp_gallery)
        add_image(sample_images["base"], temp_gallery, force_base=True)
        result = add_image(sample_images["base"], temp_gallery)
        assert result.success
        assert result.data["status"] == "already_exists"

    def test_add_without_init(self, temp_gallery, sample_images):
        result = add_image(sample_images["base"], temp_gallery)
        assert not result.success
        assert "not initialized" in result.error

    def test_add_nonexistent_image(self, temp_gallery):
        init_gallery(temp_gallery)
        result = add_image("/nonexistent/image.png", temp_gallery)
        assert not result.success
        assert "not found" in result.error


class TestRemoveImage:
    def test_remove_image(self, initialized_gallery, sample_images):
        from kantu.api import add_image, get_image_info

        result = add_image(sample_images["similar"], initialized_gallery)
        image_id = result.data["id"]
        remove_result = remove_image(image_id, initialized_gallery)
        assert remove_result.success
        info = get_image_info(image_id, initialized_gallery)
        assert not info.success

    def test_remove_nonexistent(self, initialized_gallery):
        result = remove_image("nonexistent123", initialized_gallery)
        assert not result.success


class TestGetImageInfo:
    def test_get_info(self, initialized_gallery):
        images = list_images(initialized_gallery)
        if images.success and images.data["images"]:
            image_id = images.data["images"][0]["id"]
            info = get_image_info(image_id, initialized_gallery)
            assert info.success
            assert info.data["id"] == image_id

    def test_get_info_nonexistent(self, initialized_gallery):
        result = get_image_info("nonexistent", initialized_gallery)
        assert not result.success


class TestListImages:
    def test_list_empty(self, temp_gallery):
        init_gallery(temp_gallery)
        result = list_images(temp_gallery)
        assert result.success
        assert result.data["total"] == 0

    def test_list_with_images(self, initialized_gallery):
        result = list_images(initialized_gallery)
        assert result.success
        assert result.data["total"] >= 1

    def test_list_pagination(self, initialized_gallery):
        result = list_images(initialized_gallery, limit=1, offset=0)
        assert result.success
        assert len(result.data["images"]) <= 1


class TestExportImage:
    def test_export_image(self, initialized_gallery, temp_gallery):
        import os

        images = list_images(initialized_gallery)
        if images.success and images.data["images"]:
            image_id = images.data["images"][0]["id"]
            output_path = os.path.join(temp_gallery, "exported.png")
            result = export_image(image_id, output_path, initialized_gallery)
            assert result.success
            assert os.path.exists(output_path)


class TestFindSimilar:
    def test_find_similar(self, initialized_gallery, sample_images):
        result = find_similar(sample_images["similar"], initialized_gallery)
        assert result.success
        assert "phash" in result.data
        assert "similar_images" in result.data


class TestGetGalleryStats:
    def test_stats(self, initialized_gallery):
        result = get_gallery_stats(initialized_gallery)
        assert result.success
        assert "total_images" in result.data
        assert "base_images" in result.data
        assert "delta_images" in result.data


class TestSetConfig:
    def test_set_config(self, initialized_gallery):
        result = set_config(
            initialized_gallery,
            similarity_threshold=0.9,
            min_delta_ratio=0.3,
        )
        assert result.success
        assert "similarity_threshold" in result.data["changes"]



"""Tests for KanTu core functionality."""

from pathlib import Path

from kantu.core import ImageRecord, KanTuCore


class TestKanTuCore:
    def test_init_gallery(self, temp_gallery):
        core = KanTuCore(temp_gallery)
        assert not core.is_initialized()
        core.init_gallery()
        assert core.is_initialized()
        assert core.kantu_path.exists()
        assert core.db_path.exists()

    def test_compute_hash(self, temp_gallery, sample_images):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        hash1 = core.compute_hash(sample_images["base"])
        hash2 = core.compute_hash(sample_images["similar"])
        assert isinstance(hash1, str)
        assert len(hash1) == 64
        distance = core.hamming_distance(hash1, hash2)
        assert distance <= 30

    def test_compute_id(self, temp_gallery, sample_images):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        id1 = core.compute_id(sample_images["base"])
        id2 = core.compute_id(sample_images["similar"])
        assert isinstance(id1, str)
        assert len(id1) == 16
        assert id1 != id2

    def test_compute_ssim(self, temp_gallery, sample_images):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        ssim_score = core.compute_ssim(sample_images["base"], sample_images["similar"])
        assert 0 <= ssim_score <= 1
        assert ssim_score > 0.9
        ssim_diff = core.compute_ssim(sample_images["base"], sample_images["different"])
        assert ssim_diff < ssim_score

    def test_hamming_distance(self, temp_gallery):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        hash1 = "a" * 64
        hash2 = "a" * 63 + "b"
        distance = core.hamming_distance(hash1, hash2)
        assert isinstance(int(distance), int)
        assert distance >= 0

    def test_pixel_delta(self, temp_gallery, sample_images):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        delta, target_size = core.compute_pixel_delta(
            sample_images["base"], sample_images["similar"]
        )
        assert delta.shape == (100, 100, 3)
        assert target_size == (100, 100)

    def test_save_and_load_delta(self, temp_gallery, sample_images):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        import numpy as np

        delta = np.random.randint(-128, 127, (100, 100, 3), dtype=np.int16)
        delta_path = core.save_delta(delta, "test_delta")
        assert Path(delta_path).exists()
        loaded = core.load_delta("test_delta")
        np.testing.assert_array_equal(delta, loaded)

    def test_add_and_get_image_record(self, temp_gallery):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        record = ImageRecord(
            id="test123",
            phash="a" * 64,
            original_path="/test/image.png",
            is_base=True,
            width=100,
            height=100,
            file_size=1000,
        )
        core.add_image_record(record)
        retrieved = core.get_image_record("test123")
        assert retrieved is not None
        assert retrieved["id"] == "test123"
        assert retrieved["is_base"] == 1

    def test_list_all_images(self, temp_gallery):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        for i in range(3):
            record = ImageRecord(
                id=f"test{i}",
                phash=f"{i}" * 64,
                original_path=f"/test/image{i}.png",
                is_base=True,
                width=100,
                height=100,
                file_size=1000,
            )
            core.add_image_record(record)
        images = core.list_all_images()
        assert len(images) == 3

    def test_gallery_stats(self, temp_gallery):
        core = KanTuCore(temp_gallery)
        core.init_gallery()
        stats = core.get_gallery_stats()
        assert stats["total_images"] == 0
        assert stats["base_images"] == 0
        record = ImageRecord(
            id="test123",
            phash="a" * 64,
            original_path="/test/image.png",
            is_base=True,
            width=100,
            height=100,
            file_size=1000,
        )
        core.add_image_record(record)
        stats = core.get_gallery_stats()
        assert stats["total_images"] == 1
        assert stats["base_images"] == 1



"""Tests for KanTu CLI."""

import subprocess
import sys
from pathlib import Path


class TestCLI:
    def test_version(self):
        result = subprocess.run(
            [sys.executable, "-m", "kantu", "--version"],
            capture_output=True,
            text=True,
        )
        assert result.returncode == 0
        assert "kantu" in result.stdout

    def test_help(self):
        result = subprocess.run(
            [sys.executable, "-m", "kantu", "--help"],
            capture_output=True,
            text=True,
        )
        assert result.returncode == 0
        assert "init" in result.stdout
        assert "add" in result.stdout
        assert "list" in result.stdout

    def test_init(self, temp_gallery):
        result = subprocess.run(
            [sys.executable, "-m", "kantu", "init", temp_gallery],
            capture_output=True,
            text=True,
        )
        assert result.returncode == 0
        assert (Path(temp_gallery) / ".kantu").exists()

    def test_list_empty(self, temp_gallery):
        subprocess.run(
            [sys.executable, "-m", "kantu", "init", temp_gallery],
            capture_output=True,
        )
        result = subprocess.run(
            [sys.executable, "-m", "kantu", "list", "-g", temp_gallery],
            capture_output=True,
            text=True,
        )
        assert result.returncode == 0
        assert "Total images: 0" in result.stdout

    def test_stats(self, temp_gallery):
        subprocess.run(
            [sys.executable, "-m", "kantu", "init", temp_gallery],
            capture_output=True,
        )
        result = subprocess.run(
            [sys.executable, "-m", "kantu", "stats", "-g", temp_gallery],
            capture_output=True,
            text=True,
        )
        assert result.returncode == 0
        assert "Gallery Statistics" in result.stdout



"""Core functionality for KanTu: image hashing, similarity detection, and delta encoding."""

import hashlib
import json
import os
import sqlite3
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional

import imagehash
import numpy as np
from PIL import Image
from skimage.metrics import structural_similarity as ssim


class DeltaType(Enum):
    NONE = "none"
    PIXEL = "pixel"
    TRANSFORM = "transform"


@dataclass
class ImageRecord:
    id: str
    phash: str
    original_path: str
    is_base: bool
    base_id: Optional[str] = None
    delta_type: DeltaType = DeltaType.NONE
    delta_path: Optional[str] = None
    width: int = 0
    height: int = 0
    file_size: int = 0
    similarity_score: float = 1.0


@dataclass
class GalleryConfig:
    similarity_threshold: float = 0.85
    min_delta_ratio: float = 0.5
    hash_size: int = 16
    max_hamming_distance: int = 10


class KanTuCore:
    KANTU_DIR = ".kantu"
    OBJECTS_DIR = "objects"
    BASE_DIR = "base"
    DELTA_DIR = "delta"
    DB_FILE = "gallery.db"
    CONFIG_FILE = "config.json"

    def __init__(self, gallery_path: str):
        self.gallery_path = Path(gallery_path).resolve()
        self.kantu_path = self.gallery_path / self.KANTU_DIR
        self.objects_path = self.kantu_path / self.OBJECTS_DIR
        self.base_path = self.objects_path / self.BASE_DIR
        self.delta_path = self.objects_path / self.DELTA_DIR
        self.db_path = self.kantu_path / self.DB_FILE
        self.config_path = self.kantu_path / self.CONFIG_FILE
        self.config = GalleryConfig()
        self._conn: Optional[sqlite3.Connection] = None

    def is_initialized(self) -> bool:
        return self.kantu_path.exists() and self.db_path.exists()

    def init_gallery(self) -> bool:
        if self.is_initialized():
            return True
        self.kantu_path.mkdir(parents=True, exist_ok=True)
        self.objects_path.mkdir(parents=True, exist_ok=True)
        self.base_path.mkdir(parents=True, exist_ok=True)
        self.delta_path.mkdir(parents=True, exist_ok=True)
        self._init_database()
        self._save_config()
        return True

    def _init_database(self) -> None:
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS images (
                id TEXT PRIMARY KEY,
                phash TEXT NOT NULL,
                original_path TEXT,
                is_base INTEGER NOT NULL,
                base_id TEXT,
                delta_type TEXT DEFAULT 'none',
                delta_path TEXT,
                width INTEGER,
                height INTEGER,
                file_size INTEGER,
                similarity_score REAL,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            )
        """)
        cursor.execute("""
            CREATE INDEX IF NOT EXISTS idx_phash ON images(phash)
        """)
        cursor.execute("""
            CREATE INDEX IF NOT EXISTS idx_base_id ON images(base_id)
        """)
        conn.commit()

    def _get_connection(self) -> sqlite3.Connection:
        if self._conn is None:
            self._conn = sqlite3.connect(str(self.db_path))
            self._conn.row_factory = sqlite3.Row
        return self._conn

    def _save_config(self) -> None:
        config_data = {
            "similarity_threshold": self.config.similarity_threshold,
            "min_delta_ratio": self.config.min_delta_ratio,
            "hash_size": self.config.hash_size,
            "max_hamming_distance": self.config.max_hamming_distance,
        }
        with open(self.config_path, "w") as f:
            json.dump(config_data, f, indent=2)

    def _load_config(self) -> None:
        if self.config_path.exists():
            with open(self.config_path) as f:
                config_data = json.load(f)
            self.config.similarity_threshold = config_data.get("similarity_threshold", 0.85)
            self.config.min_delta_ratio = config_data.get("min_delta_ratio", 0.5)
            self.config.hash_size = config_data.get("hash_size", 16)
            self.config.max_hamming_distance = config_data.get("max_hamming_distance", 10)

    def compute_hash(self, image_path: str) -> str:
        with Image.open(image_path) as img:
            if img.mode != "RGB":
                img = img.convert("RGB")
            img_hash = imagehash.phash(img, hash_size=self.config.hash_size)
            return str(img_hash)

    def compute_id(self, image_path: str) -> str:
        hasher = hashlib.sha256()
        with open(image_path, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                hasher.update(chunk)
        return hasher.hexdigest()[:16]

    def compute_ssim(self, img1_path: str, img2_path: str) -> float:
        with Image.open(img1_path) as img1:
            if img1.mode != "RGB":
                img1 = img1.convert("RGB")
            arr1 = np.array(img1)
        with Image.open(img2_path) as img2:
            if img2.mode != "RGB":
                img2 = img2.convert("RGB")
            arr2 = np.array(img2)
        min_h = min(arr1.shape[0], arr2.shape[0])
        min_w = min(arr1.shape[1], arr2.shape[1])
        arr1 = arr1[:min_h, :min_w]
        arr2 = arr2[:min_h, :min_w]
        if arr1.shape[2] == 3:
            scores = []
            for i in range(3):
                score = ssim(arr1[:, :, i], arr2[:, :, i], data_range=255)
                scores.append(score)
            return float(np.mean(scores))
        else:
            return float(ssim(arr1, arr2, data_range=255))

    def hamming_distance(self, hash1: str, hash2: str) -> int:
        return imagehash.hex_to_hash(hash1) - imagehash.hex_to_hash(hash2)

    def find_similar_images(self, phash: str, threshold: Optional[int] = None) -> list[dict]:
        if threshold is None:
            threshold = self.config.max_hamming_distance
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM images WHERE is_base = 1")
        candidates = []
        for row in cursor.fetchall():
            stored_hash = row["phash"]
            distance = self.hamming_distance(phash, stored_hash)
            if distance <= threshold:
                candidates.append(
                    {
                        "id": row["id"],
                        "phash": stored_hash,
                        "distance": distance,
                        "original_path": row["original_path"],
                    }
                )
        candidates.sort(key=lambda x: x["distance"])
        return candidates

    def compute_pixel_delta(
        self, img1_path: str, img2_path: str
    ) -> tuple[np.ndarray, tuple[int, int]]:
        with Image.open(img1_path) as img1:
            if img1.mode != "RGB":
                img1 = img1.convert("RGB")
            arr1 = np.array(img1, dtype=np.int16)
        with Image.open(img2_path) as img2:
            if img2.mode != "RGB":
                img2 = img2.convert("RGB")
            arr2 = np.array(img2, dtype=np.int16)
        max_h = max(arr1.shape[0], arr2.shape[0])
        max_w = max(arr1.shape[1], arr2.shape[1])
        padded1 = np.zeros((max_h, max_w, 3), dtype=np.int16)
        padded2 = np.zeros((max_h, max_w, 3), dtype=np.int16)
        padded1[: arr1.shape[0], : arr1.shape[1]] = arr1
        padded2[: arr2.shape[0], : arr2.shape[1]] = arr2
        delta = padded2 - padded1
        return delta.astype(np.int16), (arr2.shape[0], arr2.shape[1])

    def apply_pixel_delta(
        self, base_path: str, delta: np.ndarray, target_size: tuple[int, int]
    ) -> np.ndarray:
        with Image.open(base_path) as img:
            if img.mode != "RGB":
                img = img.convert("RGB")
            base_arr = np.array(img, dtype=np.int16)
        max_h = max(base_arr.shape[0], delta.shape[0])
        max_w = max(base_arr.shape[1], delta.shape[1])
        padded_base = np.zeros((max_h, max_w, 3), dtype=np.int16)
        padded_base[: base_arr.shape[0], : base_arr.shape[1]] = base_arr
        result = padded_base + delta
        result = np.clip(result, 0, 255).astype(np.uint8)
        return result[: target_size[0], : target_size[1]]

    def save_delta(self, delta: np.ndarray, delta_id: str) -> str:
        delta_file = self.delta_path / f"{delta_id}.npz"
        np.savez_compressed(str(delta_file), delta=delta)
        return str(delta_file)

    def load_delta(self, delta_id: str) -> np.ndarray:
        delta_file = self.delta_path / f"{delta_id}.npz"
        data = np.load(str(delta_file))
        return data["delta"]

    def save_base_image(self, image_path: str, image_id: str) -> str:
        base_file = self.base_path / f"{image_id}.png"
        with Image.open(image_path) as img:
            if img.mode != "RGB":
                img = img.convert("RGB")
            img.save(base_file, "PNG")
        return str(base_file)

    def get_image_dimensions(self, image_path: str) -> tuple[int, int]:
        with Image.open(image_path) as img:
            return img.size

    def get_file_size(self, path: str) -> int:
        return os.path.getsize(path)

    def add_image_record(self, record: ImageRecord) -> None:
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute(
            """
            INSERT OR REPLACE INTO images
            (id, phash, original_path, is_base, base_id, delta_type,
             delta_path, width, height, file_size, similarity_score)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            """,
            (
                record.id,
                record.phash,
                record.original_path,
                1 if record.is_base else 0,
                record.base_id,
                record.delta_type.value,
                record.delta_path,
                record.width,
                record.height,
                record.file_size,
                record.similarity_score,
            ),
        )
        conn.commit()

    def get_image_record(self, image_id: str) -> Optional[dict]:
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM images WHERE id = ?", (image_id,))
        row = cursor.fetchone()
        if row:
            return dict(row)
        return None

    def list_all_images(self) -> list[dict]:
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM images ORDER BY created_at DESC")
        return [dict(row) for row in cursor.fetchall()]

    def delete_image_record(self, image_id: str) -> bool:
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM images WHERE id = ?", (image_id,))
        row = cursor.fetchone()
        if not row:
            return False
        delta_path = row["delta_path"]
        if delta_path and os.path.exists(delta_path):
            os.remove(delta_path)
        cursor.execute("DELETE FROM images WHERE id = ?", (image_id,))
        conn.commit()
        return True

    def get_gallery_stats(self) -> dict:
        conn = self._get_connection()
        cursor = conn.cursor()
        cursor.execute("SELECT COUNT(*) FROM images")
        total_images = cursor.fetchone()[0]
        cursor.execute("SELECT COUNT(*) FROM images WHERE is_base = 1")
        base_images = cursor.fetchone()[0]
        cursor.execute("SELECT COUNT(*) FROM images WHERE is_base = 0")
        delta_images = cursor.fetchone()[0]
        cursor.execute("SELECT SUM(file_size) FROM images WHERE is_base = 1")
        base_size = cursor.fetchone()[0] or 0
        cursor.execute(
            "SELECT SUM(file_size) FROM images WHERE is_base = 0 AND delta_path IS NOT NULL"
        )
        delta_size = cursor.fetchone()[0] or 0
        cursor.execute("SELECT original_path, file_size FROM images WHERE is_base = 0")
        original_total = 0
        for row in cursor.fetchall():
            if row["original_path"] and os.path.exists(row["original_path"]):
                original_total += os.path.getsize(row["original_path"])
        savings = 0
        if original_total > 0:
            savings = 1 - (base_size + delta_size) / original_total
        return {
            "total_images": total_images,
            "base_images": base_images,
            "delta_images": delta_images,
            "base_size": base_size,
            "delta_size": delta_size,
            "total_size": base_size + delta_size,
            "original_size": original_total,
            "savings_ratio": savings,
        }

    def reconstruct_image(self, image_id: str) -> Optional[np.ndarray]:
        record = self.get_image_record(image_id)
        if not record:
            return None
        if record["is_base"]:
            base_file = self.base_path / f"{image_id}.png"
            if base_file.exists():
                with Image.open(base_file) as img:
                    return np.array(img)
            return None
        base_id = record["base_id"]
        if not base_id:
            return None
        base_record = self.get_image_record(base_id)
        if not base_record:
            return None
        base_file = self.base_path / f"{base_id}.png"
        if not base_file.exists():
            return None
        delta_id = Path(record["delta_path"]).stem if record["delta_path"] else image_id
        delta = self.load_delta(delta_id)
        target_size = (record["height"], record["width"])
        return self.apply_pixel_delta(str(base_file), delta, target_size)

    def export_image(self, image_id: str, output_path: str) -> bool:
        arr = self.reconstruct_image(image_id)
        if arr is None:
            return False
        img = Image.fromarray(arr)
        img.save(output_path)
        return True

    def close(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None
